function love.load()
  retroinput = require "retroinput"
  retroinput:init()
  retroinput.pressed = pressed
  
  tetris = require "yeetris"
  
  love.graphics.setDefaultFilter("nearest")
  graphics = {
    ui = love.graphics.newImage("ui.png"),
    tiles = {
      love.graphics.newImage("tile1.png"),
      love.graphics.newImage("tile2.png"),
      love.graphics.newImage("tile3.png")
    },
    font = love.graphics.newFont("font.ttf", 20),
    playfieldunit = 20,
    colors = {
      {0, 1, 0}, {0.5, 1, 0}, {1, 1, 0}, {1, 0.5, 0}, {1, 0, 0}
    }
  }
  love.graphics.setFont(graphics.font)
  
  state = "menu" -- "menu" or "game"
  menuoptions = {
    "resume",
    "new: %d",
    "hide",
    "quit"
  }
  menuoption = 2
  hidden = false
  starttypes = {
    9, 12, 15, 18, 19, 29
  }
  starttype = 4
  
  sessionstats = {
    top = 0,
    sectop = 0,
    maxes = 0,
    topavg = 0
  }
  postmaxscores = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
  
  local w,h = love.window.getDesktopDimensions()
  love.window.setPosition(0, h - 420)
end

function love.update(dt)
  retroinput:update()
  
  if hidden then return end
  
  love.timer.sleep(tetris:update(dt, retroinput:getAxis("left", "right"), retroinput:getAxis("a", "b")))
end

function love.draw()
  --reset colors
  love.graphics.setColor(1, 1, 1)
  love.graphics.setBackgroundColor(0, 0, 0)
  --calculate the scale of the UI
  local uiscale = love.graphics.getWidth() / graphics.ui:getWidth()
  --draw the UI frame
  love.graphics.draw(graphics.ui, 0, 0, 0, uiscale)
  --draw the stats
  if state == "game" then
    local stats = tetris:getStatsTable()
    love.graphics.print("lines", 113 * uiscale, 54 * uiscale)
    love.graphics.print(string.format("%04d", stats.lines or 0), 113 * uiscale, 66 * uiscale)
    local scorestr = string.format("%07d", stats.score or 0)
    if scorestr:len() > 6 then
      local n = tonumber(scorestr:sub(1,-7)) - 9
      if n > postmaxscores:len() then n = postmaxscores:len() end
      if n > 0 then scorestr = postmaxscores:sub(n,n) .. scorestr:sub(1,-7) end
    end
    love.graphics.print("score", 113 * uiscale, 86 * uiscale)
    love.graphics.print(scorestr:sub(1, 6), 113 * uiscale, 98 * uiscale)
    love.graphics.print("level", 113 * uiscale, 118 * uiscale)
    love.graphics.print(string.format("%03d", stats.level or 0), 113 * uiscale, 130 * uiscale)
  elseif state == "menu" then
    local stats = sessionstats
    local topstr = string.format("%07d", stats.top or 0)
    if topstr:len() > 6 then
      local n = tonumber(topstr:sub(1,-7)) - 9
      if n > postmaxscores:len() then n = postmaxscores:len() end
      if n > 0 then topstr = postmaxscores:sub(n,n) .. topstr:sub(1,-7) end
    end
    love.graphics.print("top", 113 * uiscale, 54 * uiscale)
    love.graphics.print(topstr:sub(1, 6), 113 * uiscale, 66 * uiscale)
    love.graphics.print("maxes", 113 * uiscale, 86 * uiscale)
    love.graphics.print(string.format("%03d", stats.maxes or 0), 113 * uiscale, 98 * uiscale)
    local topavgstr = string.format("%07d", stats.topavg or 0)
    if topavgstr:len() > 6 then
      local n = tonumber(topavgstr:sub(1,-7)) - 9
      if n > postmaxscores:len() then n = postmaxscores:len() end
      if n > 0 then topavgstr = postmaxscores:sub(n,n) .. topstr:sub(1,-7) end
    end
    love.graphics.print("topavg", 113 * uiscale, 118 * uiscale)
    love.graphics.print(topavgstr:sub(1, 6), 113 * uiscale, 130 * uiscale)
  end
  --calculate color from average board height
  local color = {0.3, 0.3, 1}
  if tetris:getStats() then
  local avgh = 0
  for x=0,9 do 
    local h = false
    for y=0,19 do
      if tetris:hasTile(x, y) and not h then
        h = 20 - y
        avgh = avgh + h
      end
    end
  end
  avgh = avgh / 10
  if avgh > 0 then
    local m = 20 / #graphics.colors
    local setcol = false
    for i,col in ipairs(graphics.colors) do
      if not setcol and avgh < i * m then
        color = col
      end
    end
  end
  end
  --draw the playfield
  local pox, poy, unit = 20 * uiscale, 0 * uiscale, graphics.playfieldunit
  if state == "game" then
    local playfield = tetris:getPlayfield()
    if playfield then for i,tile in ipairs(playfield) do
      love.graphics.setColor(color)
      love.graphics.rectangle("fill", tile.x * unit + pox, tile.y * unit + poy, unit, unit)
      love.graphics.setColor(1, 1, 1)
      love.graphics.draw(graphics.tiles[tile.tiletype], tile.x * unit + pox, tile.y * unit + poy, 0, unit / graphics.tiles[tile.tiletype]:getWidth())
    end end
    local nxt,tiletype = tetris:getNextPiece()
    if nxt and tiletype then for i=1,4 do
      love.graphics.setColor(color)
      love.graphics.rectangle("fill", (113 + (nxt[i][1] + 2) * 7) * uiscale, (6 + (nxt[i][2] + 2) * 7) * uiscale, 7 * uiscale, 7 * uiscale)
      love.graphics.setColor(1, 1, 1)
      love.graphics.draw(graphics.tiles[tiletype], (113+ (nxt[i][1] + 2) * 7) * uiscale, (6 + (nxt[i][2] + 2) * 7) * uiscale, 0, 7 * uiscale / graphics.tiles[tiletype]:getWidth())
    end end
  elseif state == "menu" then
    local spacing, lineheight = 4, 8
    love.graphics.print("yeetris", 8 * uiscale + pox, 16 * uiscale + poy, 0, 1.5, 1.5)
    for i,opt in ipairs(menuoptions) do
      love.graphics.print(string.format(opt, starttypes[starttype]), 8 * uiscale + pox, (12*i + 24) * uiscale + poy)
      if i == menuoption then
        love.graphics.print(">", pox + 2, (12*i + 24) * uiscale + poy)
      end
    end
  end
end

function pressed(key)
  if state == "menu" and not hidden then
    if key == "up" then
      menuoption = menuoption - 1
    elseif key == "down" then
      menuoption = menuoption + 1
    end
    if menuoption < 1 then menuoption = #menuoptions end
    if menuoption > #menuoptions then menuoption = 1 end
    if menuoption == 2 then
      if key == "left" then
        starttype = starttype - 1
      elseif key == "right" then
        starttype = starttype + 1
      end
      if starttype < 1 then starttype = #starttypes end
      if starttype > #starttypes then starttype = 1 end
    end
  end
  if key == "start" and retroinput:isDown("select") then
    hidden = not hidden
    if hidden then
      love.window.setMode(1, 1, {borderless = true})
      love.window.setPosition(0, h - 420)
    else
      love.window.setMode(420, 420, {borderless = true})
      love.window.setPosition(0, h - 420)
    end
  elseif key == "start" and not hidden then
    if state == "menu" then
      if menuoption == 1 and tetris:getStats() then
        state = "game"
      elseif menuoption == 2 then
        tetris:newGame(starttypes[starttype])
        state = "game"
      elseif menuoption == 3 then
        hidden = true
        love.window.setMode(1, 1, {borderless = true})
      elseif menuoption == 4 then
        love.event.quit()
      end
    elseif state == "game" then
      if tetris:isTopout() then
        local stats = tetris:getStatsTable()
        if stats.score > sessionstats.top then
          sessionstats.top = stats.score
        elseif stats.score > sessionstats.sectop then
          sessionstats.sectop = stats.score
        end
        sessionstats.topavg = (sessionstats.top + sessionstats.sectop) / 2
        if stats.score >= 999999 then
          sessionstats.maxes = sessionstats.maxes + 1
        end
      end
      state = "menu"
    end
  end
end