--
-- retroinput
--
-- Copyright (c) 2021 mk8
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy of
-- this software and associated documentation files (the "Software"), to deal in
-- the Software without restriction, including without limitation the rights to
-- use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
-- of the Software, and to permit persons to whom the Software is furnished to do
-- so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in all
-- copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
-- SOFTWARE.
--

local function split(str, sep) sep=sep or '%s' local t={}  for field,s in string.gmatch(str, "([^"..sep.."]*)("..sep.."?)") do table.insert(t,field)  if s=="" then return t end end end

local retroinput = {init = false}

function retroinput:init(cfgname, defkeys, defctrls)
  local newline = "\n"
  if love.system.getOS() == "Windows" then newline = "\r\n"
  elseif love.system.getOS() == "OS X" then newline = "\r" end
  
  self.input = {}
  self.inputkeys = defkeys or {"left", "right", "up", "down", "a", "b", "start", "select"}
  
  if not love.filesystem.getInfo((cfgname or "controls.cfg")) then
    local ctrls = defctrls or {
      left = "left,a,JOY_A_1_NEG,PAD_A_LEFTX_NEG,PAD_B_DPLEFT",
      right = "right,d,JOY_A_1_POS,PAD_A_LEFTX_POS,PAD_B_DPRIGHT",
      up = "up,w,JOY_A_2_NEG,PAD_A_LEFTY_NEG,PAD_B_DPUP",
      down = "down,s,JOY_A_2_POS,PAD_A_LEFTY_POS,PAD_B_DPDOWN",
      a = "c,i,JOY_B_2,PAD_B_A",
      b = "x,u,JOY_B_1,PAD_B_B",
      ["start"] = "lshift,l,JOY_B_10,PAD_B_START",
      ["select"] = "z,k,JOY_B_9,PAD_B_BACK"
    }
    local str = ""
    for i,key in ipairs(self.inputkeys) do
      if ctrls[key] then str = str .. key .. ":" .. ctrls[key] .. newline end
    end
    love.filesystem.write((cfgname or "controls.cfg"), str)
  end
  
  self.controls = {}
  local data,other = love.filesystem.read((cfgname or "controls.cfg"))
  for i,key in ipairs(self.inputkeys) do
    for line in love.filesystem.lines((cfgname or "controls.cfg")) do
      local data = split(line, ":")
      local d_key, d_controls = data[1], data[2]:sub(1, -newline:len()+1)
      if key == d_key then self.controls[key] = d_controls end
    end
  end
  
  self.init = true
end

function retroinput:update()
  if not self.init then print("retroinput is not initialized!") return end
  
  local newinput = {}
  for i,key in ipairs(self.inputkeys) do
    local down = 0
    local buttons = split(self.controls[key], ",")
    for i,button in ipairs(buttons) do
      if #split(button, "_") > 1 then
        --joystick or gamepad
        local data = split(string.upper(button), "_")
        --1 is device type (JOY or PAD), 2 is input type (axis or button), 3 is input index, [optional] 4 is input value
        for i,joystick in ipairs(love.joystick.getJoysticks()) do
          if joystick:isGamepad() and data[1] == "PAD" then
            if data[2] == "A" then
              local val = joystick:getGamepadAxis(string.lower(data[3]))
              if val > 0 and data[4] == "POS" then
                down = 1
              elseif val < 0 and data[4] == "NEG" then
                down = 1
              end
            elseif data[2] == "B" then
              if joystick:isGamepadDown(string.lower(data[3])) then
                down = 1
              end
            end
          elseif data[1] == "JOY" then
            if data[2] == "A" then
              local val = joystick:getAxis(tonumber(data[3]))
              if val > 0 and data[4] == "POS" then
                down = 1
              elseif val < 0 and data[4] == "NEG" then
                down = 1
              end
            elseif data[2] == "B" then
              if joystick:isDown(tonumber(data[3])) then
                down = 1
              end
            end
          end
        end
      else
        --keyboard
        if love.keyboard.isDown(button) then down = 1 end
      end
    end
    
    if down == 1 and self.input[key] == 0 then
      --press callback
      if self.pressed then if type(self.pressed) == "function" then
        self.pressed(key)
      end end
    elseif down == 1 and self.input[key] == 1 then
      --hold callback
      if self.held then if type(self.held) == "function" then
        self.held(key)
      end end
    elseif down == 0 and self.input[key] == 1 then
      --release callback
      if self.released then if type(self.released) == "function" then
        self.released(key)
      end end
    end
    
    newinput[key] = down
  end
  
  self.input = newinput
end

function retroinput:getInputs()
  if not self.init then print("retroinput is not initialized!") return end
  
  return self.input
end

function retroinput:isDown(key)
  if not self.init then print("retroinput is not initialized!") return end
  
  return self.input[key] == 1
end

function retroinput:getAxis(neg, pos)
  if not self.init then print("retroinput is not initialized!") return end
  
  return (self.input[pos] - self.input[neg])
end

return retroinput