local tetris = {}

tetris.pieces = { -- { {x1, y1}, {x2, y2}, {x3, y3}, {x4, y4} }
  { { -1,  0 }, {  0,  0 }, {  1,  0 }, {  0, -1 }, },  --  1: T up
  { {  0, -1 }, {  0,  0 }, {  1,  0 }, {  0,  1 }, },  --  2: T right
  { { -1,  0 }, {  0,  0 }, {  1,  0 }, {  0,  1 }, },  --  3: T down (spawn)
  { {  0, -1 }, { -1,  0 }, {  0,  0 }, {  0,  1 }, },  --  4: T left

  { {  0, -1 }, {  0,  0 }, { -1,  1 }, {  0,  1 }, },  --  5: J left
  { { -1, -1 }, { -1,  0 }, {  0,  0 }, {  1,  0 }, },  --  6: J up
  { {  0, -1 }, {  1, -1 }, {  0,  0 }, {  0,  1 }, },  --  7: J right
  { { -1,  0 }, {  0,  0 }, {  1,  0 }, {  1,  1 }, },  --  8: J down (spawn)

  { { -1,  0 }, {  0,  0 }, {  0,  1 }, {  1,  1 }, },  --  9: Z horizontal (spawn) 
  { {  1, -1 }, {  0,  0 }, {  1,  0 }, {  0,  1 }, },  -- 10: Z vertical

  { { -1,  0 }, {  0,  0 }, { -1,  1 }, {  0,  1 }, },  -- 11: O (spawn)

  { {  0,  0 }, {  1,  0 }, { -1,  1 }, {  0,  1 }, },  -- 12: S horizontal (spawn)
  { {  0, -1 }, {  0,  0 }, {  1,  0 }, {  1,  1 }, },  -- 13: S vertical

  { {  0, -1 }, {  0,  0 }, {  0,  1 }, {  1,  1 }, },  -- 14: L right
  { { -1,  0 }, {  0,  0 }, {  1,  0 }, { -1,  1 }, },  -- 15: L down (spawn)
  { { -1, -1 }, {  0, -1 }, {  0,  0 }, {  0,  1 }, },  -- 16: L left
  { {  1, -1 }, { -1,  0 }, {  0,  0 }, {  1,  0 }, },  -- 17: L up

  { {  0, -2 }, {  0, -1 }, {  0,  0 }, {  0,  1 }, },  -- 18: I vertical
  { { -2,  0 }, { -1,  0 }, {  0,  0 }, {  1,  0 }, }   -- 19: I horizontal (spawn)
}

tetris.spawns = {
  3, 8, 9, 11, 12, 15, 19
}

--piece generation

function tetris:newPieceSpawnId()
  return math.random(#self.spawns)
end

function tetris:newPieceId()
  return self.spawns[self:newPieceSpawnId()]
end

function tetris:newPieceSpawn(id)
  local _id = id or math.random(#self.spawns)
  return self:newPiece(self.spawns[id])
end

function tetris:newPiece(id)
  return {x=5,y=0,id=id}
end

tetris.rotations = { -- {clockwise, counterclockwise}
  {  2,  4  }, --  1: T up
  {  3,  1  }, --  2: T right
  {  4,  2  }, --  3: T down
  {  1,  3  }, --  4: T left
  
  {  6,  8  }, --  5: J left
  {  7,  5  }, --  6: J up
  {  8,  6  }, --  7: J right
  {  5,  7  }, --  8: J down
  
  { 10, 10  }, --  9: Z horizontal
  {  9,  9  }, -- 10: Z vertical
  
  { 11, 11  }, -- 11: O
  
  { 13, 13  }, -- 12: S horizontal
  { 12, 12  }, -- 13: S vetical
  
  { 15, 17  }, -- 14: L right
  { 16, 14  }, -- 15: L down
  { 17, 15  }, -- 16: L left
  { 14, 16  }, -- 17: L up
  
  { 19, 19  }, -- 18: I vertical
  { 18, 18  }  -- 19: I horizontal
}

tetris.tiles = {
  3, 3, 3, 3, 1, 1, 1, 1, 2, 2, 3, 1, 1, 2, 2, 2, 2, 3, 3
}

tetris.framerate = 60.0988

tetris.fallspeed = {
  6, 6, 6, 6, 6, 6, 6, 6, 6, 5, 5, 5, 4, 4, 4, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1
}

tetris.input = {
  dir = 0,
  das = 0,
  rot = 0,
  _dasdelay = 16,
  _dasrepeat = 6
}

function tetris:newGame(level)
  self.init = true
  
  self.level = level or 0
  self.lines = 0
  self.score = 0
  self.topout = false
  self.board = {}
  self.piece = false
  self.nextpiece = false
  self.lineclear = false
  self.fallcountdown = 0
  self.ARE = 0
end

function tetris:frame(dir, rot)
  if self.topout or not self.init then
    return
  end
  
  if not self.piece then
    self.ARE = self.ARE + 1
  end
  
  if self.lineclear then
    local typ, data = self.lineclear:update()
    local func = "clear" .. string.upper(typ:sub(1, 1)) .. string.lower(typ:sub(2))
    self[func](self, data)
    
    self.input.dir, self.input.rot = dir, rot
    return
  elseif self:getFullLines() then
    self.lineclear = self:newLineClear(self:getFullLines(), 4)
  end
  
  if not self.piece and self.ARE >= 14 then
    self.piece = self:newPiece(self.nextpiece or self:newPieceId())
    self.nextpiece = self:newPieceId()
    self.ARE = 0
    --if piece collides on spawn, topout
    if self:checkPiece(self.piece.x, self.piece.y, self.piece.id) then
      self.topout = true
    end
  elseif self.piece then
    --piece movement
    if dir ~= 0 and self.input.dir == 0 then
      --tap
      self.input.das = 0
      if not self:checkPiece(self.piece.x + dir, self.piece.y, self.piece.id) then
        self.piece.x = self.piece.x + dir
      end
    elseif dir ~= 0 then
      --hold
      self.input.das = self.input.das + 1
      if self.input.das >= self.input._dasdelay then
        if not self:checkPiece(self.piece.x + dir, self.piece.y, self.piece.id) then
          self.piece.x = self.piece.x + dir
          self.input.das = self.input.das - self.input._dasrepeat
        end
      end
    end
    --wall charging
    if self:checkPiece(self.piece.x + dir, self.piece.y, self.piece.id) then
      self.input.das = self.input._dasdelay
    end
    --piece rotation
    if rot ~= 0 and rot ~= self.input.rot then
      --clockwise
      if rot > 0 and not self:checkPiece(self.piece.x, self.piece.y, self.rotations[self.piece.id][1]) then
        self.piece.id = self.rotations[self.piece.id][1]
      --counterclockwise
      elseif rot < 0 and not self:checkPiece(self.piece.x, self.piece.y, self.rotations[self.piece.id][2]) then
        self.piece.id = self.rotations[self.piece.id][2]
      end
    end
    --falling
    self.fallcountdown = self.fallcountdown + 1
    local fallspeedindex = self.level + 1
    if fallspeedindex > #self.fallspeed then fallspeedindex = self.fallspeed end
    if self.fallcountdown >= self.fallspeed[fallspeedindex] then 
      if not self:checkPiece(self.piece.x, self.piece.y + 1, self.piece.id) then
        --move piece down
        self.piece.y = self.piece.y + 1
      elseif self:checkPiece(self.piece.x, self.piece.y + 1, self.piece.id) then
        --lock the piece and delete it
        self:lockPiece(self.piece.x, self.piece.y, self.piece.id, self.tiles[self.piece.id])
        self.piece = false
      end
      self.fallcountdown = self.fallcountdown - self.fallspeed[fallspeedindex]
    end
  end
  
  --update the input
  self.input.dir, self.input.rot = dir, rot
  return
end

function tetris:update(dt, dir, rot)
  local frames = math.ceil(dt * self.framerate)
  local wait = frames / self.framerate - dt
  
  for i=1,frames do
    self:frame(dir or 0, rot or 0)
  end
  
  return wait
end

--CLASSES

--line clear
function tetris:newLineClear(lines, framespeed)
  local lineclear = {lines = lines, framespeed = framespeed, countdown = 0, state = 0}
  function lineclear:update()
    self.countdown = self.countdown + 1
    if self.countdown >= self.framespeed then
      self.state = self.state + 1
      if self.state > 0 then
        local cleartiles = {}
        for i=1,#self.lines do
          table.insert(cleartiles, {x = 4 + self.state, y = self.lines[i]})
          table.insert(cleartiles, {x = 5 - self.state, y = self.lines[i]})
        end
        if self.state >= 5 then
          local lines = self.lines
          self = false
          return "lines", lines
        end
        return "tiles", cleartiles
      end
    end
  end
end


--data retrieving functions

function tetris:hasTile(x, y)
  if x < 0 or x > 9 or y < -2 or y > 19 then
    print("{" .. tostring(x) .. "," .. tostring(y) .. "} is out of bounds")
    return true
  elseif x >= 0 and x <= 9 and y >= 0 and y <= 19 then
    for i,tile in ipairs(self.board) do
      if tile.x == x and tile.y == y then
        return true
      end
    end
  end
  return false
end

function tetris:clearTile(x, y)
  local toclear = {}
  local i = 1
  while i <= #self.board do
    if self.board[i].x == x and self.board[i].y == y then
      table.remove(self.board, i)
    else
      i=i+1
    end
  end
end

function tetris:clearTiles(tiles)
  for i,tile in ipairs(tiles) do
    self:clearTile(tile.x, tile.y)
  end
end

function tetris:isLineFull(line)
  local full = true
  for x=0,9 do
    if not self:hasTile(x, line) then
      full = false
    end
  end
  return full
end

function tetris:clearLine(line)
  for i=0,9 do
    self:clearTile(i, line)
  end
end

function tetris:getFullLines()
  local lines = false
  for line=0,19 do
    if self:isLineFull(line) then
      if not lines then lines = {line}
      else table.insert(lines, line) end
    end
  end
  return lines
end

function tetris:clearLines(lines)
  for i,line in ipairs(lines) do
    self:clearLine(line)
  end
end

function tetris:checkPiece(x, y, id)
  for i=1,4 do
    if self:hasTile(x + self.pieces[id][i][1], y + self.pieces[id][i][2]) then
      print("board collision on coordinates {" .. tostring(x+self.pieces[id][i][1]) .. "," .. tostring(y+self.pieces[id][i][2]) .. "} with piece {x:" ..  tostring(x) .. ",y:" .. tostring(y) .. ",id:" .. tostring(id) .. "}")
      return true
    end
  end
  return false
end

function tetris:lockPiece(x, y, id)
  for i=1,4 do
    table.insert(self.board, {x = x + self.pieces[id][i][1], y = y + self.pieces[id][i][2], tiletype = self.tiles[id]})
  end
  return false
end

function tetris:getPlayfield()
  local board = self.board
  if self.piece then for i=1,4 do
    table.insert(board, {x = self.piece.x + self.pieces[self.piece.id][i][1], y = self.piece.y + self.pieces[self.piece.id][i][2], tiletype = self.tiles[self.piece.id]})
  end end
  return board
end

function tetris:getNextPiece()
  if self.nextpiece then
    return self.pieces[self.nextpiece], self.tiles[self.nextpiece]
  end
end

function tetris:isTopout()
  return self.topout
end

function tetris:getStats()
  return self.lines, self.score, self.level
end

function tetris:getStatsList()
  return {self.lines, self.score, self.level}
end

function tetris:getStatsTable()
  return {lines = self.lines, score = self.score, level = self.level}
end

function tetris:getInput()
  return self.dir, self.das, self.rot
end

function tetris:getInputList()
  return {self.dir, self.das, self.rot}
end

function tetris:getInputTable()
  return {dir = self.dir, das = self.das, rot = self.rot}
end

return tetris